# -*- coding: utf8 -*-
"""
全局变量在这里定义
"""
import traceback
import sys
import os

print("加载: Globals.py")
# sys.path.append(sys.path[0])	# 添加workspace的路径(demo.py)
# sys.path.append(os.path.split(os.path.realpath(__file__))[0][0:-5])	# 添加dHydra目录的根路径(app.py)的路径

# producer
PRODUCER_HASH = dict()
PRODUCER_NAME = dict()

actionDict = dict()



# -*- coding: utf8 -*-
# 以下是自动生成的 #
# --- 导入系统配置
from dHydra.app import PRODUCER_NAME, PRODUCER_HASH
import dHydra.core.util as util
from dHydra.core.Producer import Producer
from dHydra.core.Event import Event
from dHydra.config import connection as CON
from dHydra.config import const as C
from dHydra.core.Functions import *
# --- 导入自定义配置
from .connection import *
from .const import *
from .config import *
# 以上是自动生成的 #
from datetime import datetime
import time
import requests
import websockets
import getpass
import base64,rsa,binascii
import json
import asyncio
import threading
import functools
import re


class SinaLevel2WSProducer(Producer):
	def __init__(self, name = None, username = None, pwd = None,raw = False, symbols = None, **kwargs):
		super().__init__( name=name, **kwargs )
		if (username == None):
			self.username = input('请输入新浪登录帐号：')
		else:
			self.username=username
		if (pwd == None):
			self.pwd = getpass.getpass("输入登录密码（密码不会显示在屏幕上，输入后按回车确定）:")
		else:
			self.pwd = pwd
		self.rsaPubkey = '10001'
		self.ip = util._get_public_ip()
		self.session = requests.Session()
		self.isLogin = self.login()
		self.raw = raw
		if symbols is None:
			sina = V('Sina')
			self.symbols = sina.get_symbols()
		else:
			self.symbols = symbols


	def login(self):
		self.session.get("http://finance.sina.com.cn/realstock/company/sz300204/l2.shtml")
		su = base64.b64encode(self.username.encode('utf-8'))

		preLogin = self.session.get( URL_PRELOGIN, params = PARAM_PRELOGIN( su ), headers = HEADERS_LOGIN)
		preLogin = json.loads( preLogin.text[len("sinaSSOController.preloginCallBack("):-1] )

		sp = self.get_sp( self.pwd, preLogin["pubkey"], int(preLogin["servertime"]), preLogin["nonce"] )

		self.loginResponse = self.session.post(
			URL_SSOLOGIN
		,	params = PARAM_LOGIN()
		,	data = DATA_LOGIN(
				su = su
			,	servertime = int( preLogin["servertime"] )
			,	nonce = preLogin["nonce"]
			,	rsakv = preLogin["rsakv"]
			,	sp = sp
			)
		,	headers = HEADERS_LOGIN
		)
		if (self.loginResponse.json()["retcode"]=='0'):
			print( "登录成功: %s, uid = %s" % ( self.loginResponse.json()["nick"], self.loginResponse.json()["uid"]) )

			i = 0
			for url in self.loginResponse.json()["crossDomainUrlList"]:
				req = self.session.get( url,headers = HEADERS_CROSSDOMAIN( CROSSDOMAIN_HOST[i] ) )
				i += 1
			return True
		else:
			print( "Authentication Failed..." )
			print( self.loginResponse.json() )
			return False

	# RSA2 encoding
	def get_sp(self, passwd, pubkey, servertime, nonce):
		key = rsa.PublicKey(int(pubkey, 16), int('10001', 16))
		message = str(servertime) + '\t' + str(nonce) + '\n' + str(passwd)
		passwd = rsa.encrypt(message.encode('utf-8'), key)
		return binascii.b2a_hex(passwd).decode('ascii')

	@asyncio.coroutine
	def get_ws_token(self,qlist):
		loop = asyncio.get_event_loop()
		async_req = loop.run_in_executor(None, functools.partial( self.session.get, 
			URL_WSKT_TOKEN
		,	params 	=	PARAM_WSKT_TOKEN(ip=self.ip,qlist=qlist)
		,	headers =	HEADERS_WSKT_TOKEN()
		,	verify	=	True
		) )
		req = yield from async_req
		token = req.text[45:-17]
		return token

	# 2cn_是3秒一条的Level2 10档行情
	# 2cn_symbol_0,2cn_symbol_1是逐笔数据
	# 2cn_symbol_orders是挂单数据
	# symbol_i是基本信息
	def generate_qlist(self,qlist,symbol):
		if qlist == '':
			qlist = "2cn_%s,2cn_%s_0,2cn_%s_1,%s,%s_i,2cn_%s_orders" % (symbol,symbol,symbol,symbol,symbol,symbol)
		else:
			qlist = qlist + ',' + "2cn_%s,2cn_%s_0,2cn_%s_1,%s,%s_i,2cn_%s_orders" % (symbol,symbol,symbol,symbol,symbol,symbol)
		return qlist

	def send(self, message, ws ):
		print( "> {}".format(message) )
		yield from ws.send(message)

	@asyncio.coroutine
	def create_ws(self, qlist, symbol, loop ):
		asyncio.set_event_loop(loop)
		retry = True
		while retry:
			try:
				token = yield from self.get_ws_token(qlist)
				retry = False
			except Exception as e:
				print(e)

		url_wss = 'wss://ff.sinajs.cn/wskt?token=' + token + '&list=' + qlist

		start = datetime.now()
		retry = True
		while retry:
			try:
				ws = yield from websockets.connect(url_wss)
				retry = False
			except Exception as e:
				print("Retrying..")

		# 另开一个线程每40秒更新一次token，新建一个event_loop防止这个操作阻塞websocket
		loopToken = asyncio.new_event_loop()
		tasks = [ self.renew_token(ws, qlist) ]
		renewToken = threading.Timer(40, util.thread_loop, (loopToken,tasks) )
		renewToken.start()

		while self._active:
			try:
				message = yield from ws.recv()
				# 如果要的不是原始数据
				if self.raw == False:
					message = self.ws_parse(message = message)

				event = Event(eventType = 'SinaLevel2WS', data = message)
				
				for q in self._subscriber:
					q.put(event)

			except Exception as e:
				print(e)
				ws.close()
				yield from self.create_ws(qlist = qlist,symbol = symbol,loop=loop)
	
	"""
	用于更新token的coroutine
	"""
	@asyncio.coroutine
	def renew_token(self, ws, qlist):
		while True:
			retry = True
			while retry:
				try:
					token = yield from self.get_ws_token(qlist)
					retry = False
				except:
					pass
			try:
				yield from ws.send("*"+token)
			except Exception as e:
				print( "发送token失败, 原因： {}".format(e) )
			yield from asyncio.sleep(40)

	"""
	供线程调用的开启新浪WebSocket的方法
	"""
	def start_ws(self, symbolList = None, loop = None ):
		asyncio.set_event_loop(loop)
		qlist = ''
		for symbol in symbolList:
			qlist = self.generate_qlist(qlist=qlist,symbol=symbol)

		qlist = qlist.lower()
		if loop.is_running():
			loop = asyncio.new_event_loop()
			asyncio.set_event_loop( loop )
		loop.run_until_complete( self.create_ws(qlist,symbol=symbol, loop=loop) )
		loop.close()


	"""
	用于解析Sina l2的函数
	"""
	def ws_parse(self, message):
		dataList = re.findall(r'(?:((?:2cn_)?((?:sh|sz)[\d]{6})(?:_0|_1|_orders|_i)?)(?:=)(.*)(?:\n))',message)
		result = list()
		for data in dataList:
			if (len(data[0])==12):	# quotation
				wstype = 'quotation'
			elif ( (data[0][-2:]=='_0') | (data[0][-2:]=='_1') ):
				wstype = 'deal'
			elif ( data[0][-6:]=='orders' ):
				wstype = 'orders'
			elif ( (data[0][-2:]=='_i') ):
				wstype = 'info'
			else:
				wstype = 'unknown'
			result = self.ws_parse_to_list(wstype=wstype,symbol=data[1],data=data[2],result=result)
		return result

	def ws_parse_to_list(self,wstype,symbol,data,result):
		data = data.split(',')
		if (wstype == 'deal'):
			for d in data:
				x = list()
				x.append(wstype)
				x.append(symbol)
				x.extend( d.split('|') )
				result.append(x)
		else:
			x = list()
			x.append(wstype)
			x.append(symbol)
			x.extend(data)
			result.append(x)
		return result

	def handler(self):
		# 首先从雪球获取股票列表
		symbolList = self.symbols
		# symbolList = ['SZ300204,SZ000001']
		threads = []
		# Cut symbolList
		step = 30
		symbolListSlice = [symbolList[ i : i + step] for i in range(0, len(symbolList), step)]
		for symbolList in symbolListSlice:
			loop = asyncio.new_event_loop()
			t = threading.Thread(target = self.start_ws,args=(symbolList,loop) )
			threads.append(t)
		for t in threads:
			t.setDaemon(True)
			t.start()
			print("开启线程：",t.name)
		for t in threads:
			t.join()


# -*- coding: utf8 -*-
"""
工具类
Created on 03/01/2016
@author: Wen Gu
@contact: emptyset110@gmail.com
"""
import dHydra.config.const as C
import requests
import threading
import asyncio
import math
import time
from datetime import datetime
import pandas
import os
import ntplib
from pandas import DataFrame

def _code_to_symbol(code):
	"""
		生成symbol代码标志
		@author: Jimmy Liu
		@group : waditu
		@contact: jimmysoa@sina.cn
	"""
	if code in C.INDEX_LABELS:
		return C.INDEX_LIST[code]
	else:
		if len(code) != 6 :
			return ''
		else:
			return 'sh%s'%code if code[:1] in ['5', '6', '9'] else 'sz%s'%code

def symbol_list_to_code(symbolList):
	codeList = []
	for symbol in symbolList:
		codeList.append(symbol[2:8])
	return codeList

def _get_public_ip():
	return requests.get('http://ipinfo.io/ip').text.strip()

# 用于将一个loop交给一个线程来完成一些任务
def thread_loop(loop, tasks):
	asyncio.set_event_loop(loop)
	loop.run_until_complete( asyncio.wait(tasks) )
	loop.close()

# 用于将一个list按一定步长切片，返回这个list切分后的list
def slice_list(step = None, num = None, dataList=None):
	if not ( (step is None) & (num is None) ):
		if num is not None:
			step = math.ceil(len(dataList)/num)
		return [dataList[ i : i + step] for i in range(0, len(dataList), step)]
	else:
		print("step和num不能同时为空")
		return False

# n个任务交给m个Thread完成
def threads_for_tasks(taskList):
	threads = []
	for task in taskList:
		t = threading.Thread(target = task.target, args = task.args)
		threads.append(t)
	for t in threads:
		t.start()
		print("开启线程：",t.name)
	for t in threads:
		t.join()

def symbols_to_string(symbols):
	if isinstance(symbols, list) or isinstance(symbols, set) or isinstance(symbols, tuple) or isinstance(symbols, pandas.Series):
		return ','.join(symbols)
	else:
		return symbols

def datetime_to_timestamp(dt, timeFormat = 'ms'):
	if timeFormat == 'ms':
		return int( time.mktime( dt.timetuple() )*1000 )
	elif timeFormat == 's':
		return int( time.mktime( dt.timetuple() ) )

def date_to_timestamp(date, dateFormat = '%Y-%m-%d', timeFormat = 'ms'):
	return datetime_to_timestamp( dt = datetime.strptime(date, dateFormat) , timeFormat = timeFormat)

def timestamp_to_datetime(timestamp, timeFormat='ms'):
	if timeFormat == 'ms':
		timestamp = timestamp/1000
	return datetime.strftime(timestamp)

def time_now():
	return int(time.time()*1000)

# 从国家授时中心获取时间戳
def get_network_time():
	start = time.time()
	c = ntplib.NTPClient() 
	response = c.request('pool.ntp.org') 
	ts = response.tx_time 
	return ts

def check_time(precision = 0.02):
	duration = 1.0
	while duration > precision:
		try:
			print("{}, 开始获取网络时间戳".format( time.time() ))
			start = time.time()
			networkTime = get_network_time()
			end = time.time()
			duration = end-start
		except:
			print("获取网络时间出了点小状况，正重试")
	# print("网络耗时：{}".format( duration ) )
	# print("{}, 网络时间戳".format( networkTime ) )
	# print("{}, 现在时间戳".format( time.time()) )
	difference = networkTime - (start+duration)
	print("difference = {}, (本地时间戳+difference)=网络时间戳".format(difference))
	return difference

def split_symbols(symbols):
	df = DataFrame(symbols, columns = ['s'])
	sz = list(df[ df.s > 'sz' ]["s"])
	sh = list(df[ df.s < 'sz' ]["s"])
	return [ sz, sh ]


# -*- coding: utf8 -*-
"""
主控制程序入口
"""
import traceback
import sys
import os

print("加载：app.py")

from dHydra.core.Globals import *
from dHydra.core.Functions import *

def start_action(actionList):
	for action in actionList:
		actionInstance = A(action)
		actionDict[action] = actionInstance
	for action in actionList:
		actionDict[action].start()
	for action in actionList:
		actionDict[action].join()




# -*- coding: utf8 -*-
import time
# 高速行情接口
URL_QUOTATION = lambda symbols: "http://hq.sinajs.cn/?rn=%s&list=%s" % ( int( time.time() )*1000, symbols )

SINA_QUOTE_COLUMNS = [	'name', 'open'
					,	'pre_close', 'price', 'high', 'low', 'bid', 'ask', 'volume', 'amount'
					,	'b1_v',	'b1_p', 'b2_v', 'b2_p', 'b3_v', 'b3_p', 'b4_v', 'b4_p', 'b5_v', 'b5_p'
					,	'a1_v',	'a1_p', 'a2_v', 'a2_p', 'a3_v', 'a3_p', 'a4_v', 'a4_p', 'a5_v', 'a5_p'
					,	'date', 'time', 'ms']

SINA_QUOTE_COLUMNS_2 = [	'name', 'open'
					,	'pre_close', 'price', 'high', 'low', 'bid', 'ask', 'volume', 'amount'
					,	'b1_v',	'b1_p', 'b2_v', 'b2_p', 'b3_v', 'b3_p', 'b4_v', 'b4_p', 'b5_v', 'b5_p'
					,	'a1_v',	'a1_p', 'a2_v', 'a2_p', 'a3_v', 'a3_p', 'a4_v', 'a4_p', 'a5_v', 'a5_p'
					,	'date', 'time', 'ms', 'symbol']

# 获取
URL_API_MARKET_CENTER_GETHQNODEDATA = lambda node: "http://vip.stock.finance.sina.com.cn/quotes_service/api/json_v2.php/Market_Center.getHQNodeData?num=5000&sort=symbol&asc=0&node=%s&symbol=&_s_r_a=page&page=1" % node
# node = hs_a, hs_b
# 数据格式中key缺少双引号
# [	
# 	{
# 		symbol:"sh600006"
# 	,	code:"600006"
# 	,	name:"东风汽车"
# 	,	trade:"6.210"
# 	,	pricechange:"0.000"
# 	,	changepercent:"0.000"
# 	,	buy:"0.000"
# 	,	sell:"0.000"
# 	,	settlement:"6.210"
# 	,	open:"0.000"
# 	,	high:"0.000"
# 	,	low:"0.000"
# 	,	volume:0
# 	,	amount:0
# 	,	ticktime:"10:12:31"
# 	,	per:88.336
# 	,	pb:1.94
# 	,	mktcap:1242000
# 	,	nmc:1242000
# 	,	turnoverratio:0
# 	}
# ]